#! /usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Setup.py for installing OpenTimelineIO

For more information:
- see README.md
- http://opentimeline.io
"""

import multiprocessing
import os
import sys
import platform
import subprocess
import unittest
import tempfile
import shutil

from setuptools import (
    setup,
    Extension,
    find_packages,
)

import setuptools.command.build_ext
import setuptools.command.build_py

SOURCE_DIR = os.path.abspath(os.path.dirname(__file__))

PLAT_TO_CMAKE = {
    "win32": "Win32",
    "win-amd64": "x64",
}


def _debugInstance(x):
    for a in sorted(dir(x)):
        print("%s:     %s" % (a, getattr(x, a)))


class CMakeExtension(Extension):
    def __init__(self, name):
        Extension.__init__(self, name, sources=[])


class OTIO_build_ext(setuptools.command.build_ext.build_ext):
    """
    def initialize_options(self):
        super(setuptools.command.build_ext.build_ext, self).initialize_options()
    """

    def run(self):
        # This works around the fact that we build _opentime and _otio
        # extensions as a one-shot cmake invocation. Usually we'd build each
        # separately using build_extension.
        self.announce('running OTIO build_ext', level=2)
        self.build()

    def build(self):
        self.build_temp_dir = (
            os.environ.get("OTIO_CXX_BUILD_TMP_DIR")
            or os.path.abspath(self.build_temp)
        )

        if not os.path.exists(self.build_temp_dir):
            os.makedirs(self.build_temp_dir)

        debug = (self.debug or bool(os.environ.get("OTIO_CXX_DEBUG_BUILD")))
        self.build_config = ('Debug' if debug else 'Release')

        self.cmake_preflight_check()
        self.cmake_generate()
        self.cmake_install()

    def is_windows(self):
        return platform.system() == "Windows"

    def is_mingw(self):
        return self.plat_name.startswith('mingw')

    def generate_cmake_arguments(self):
        # Use the provided build dir so setuptools will be able to locate and
        # either install to the correct location or package.
        install_dir = os.path.abspath(self.build_lib)
        if not install_dir.endswith(os.path.sep):
            install_dir += os.path.sep

        cmake_args = [
            # Python_EXECUTABLE is important as it tells CMake's FindPython
            # which Python executable to use. We absolutely want to use the
            # interpreter that was used to execute the setup.py.
            # See https://cmake.org/cmake/help/v3.20/module/FindPython.html#artifacts-specification # noqa: E501
            # Also, be careful, CMake is case sensitive ;)
            '-DPython_EXECUTABLE=' + sys.executable,
            '-DOTIO_PYTHON_INSTALL:BOOL=ON',
            '-DOTIO_CXX_INSTALL:BOOL=OFF',
            '-DOTIO_SHARED_LIBS:BOOL=OFF',
            '-DCMAKE_BUILD_TYPE=' + self.build_config,
            '-DOTIO_PYTHON_INSTALL_DIR=' + install_dir,
            # turn off the C++ tests during a Python build
            '-DBUILD_TESTING:BOOL=OFF',
            # Python modules wil be installed by setuptools.
            '-DOTIO_INSTALL_PYTHON_MODULES:BOOL=OFF',
        ]
        if self.is_windows():
            if self.is_mingw():
                cmake_args += ['-G Unix Makefiles']
            else:
                cmake_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]

        cxx_coverage = bool(os.environ.get("OTIO_CXX_COVERAGE_BUILD"))
        if cxx_coverage and not os.environ.get("OTIO_CXX_BUILD_TMP_DIR"):
            raise RuntimeError(
                "C++ code coverage requires that both OTIO_CXX_COVERAGE_BUILD=ON "
                "and OTIO_CXX_BUILD_TMP_DIR are specified as environment "
                "variables, otherwise coverage cannot be generated."
            )

        if cxx_coverage:
            cmake_args += ['-DOTIO_CXX_COVERAGE=1']

        # allow external arguments to cmake via the CMAKE_ARGS env var
        cmake_args += [
            arg for arg in os.environ.get("CMAKE_ARGS", "").split(" ")
            if arg
        ]

        return cmake_args

    def cmake_preflight_check(self):
        """
        Verify that CMake is greater or equal to the required version
        We do this so that the error message is clear if the minimum version is not met.
        """
        self.announce('running cmake check', level=2)
        # We need to run cmake --check-system-vars because it will still generate
        # a CMakeCache.txt file.
        tmpdir = tempfile.mkdtemp(dir=self.build_temp_dir)

        args = ["--check-system-vars", SOURCE_DIR] + self.generate_cmake_arguments()

        proc = subprocess.Popen(
            ["cmake"] + args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=tmpdir,
            universal_newlines=True
        )

        _, stderr = proc.communicate()
        if proc.returncode != 0:
            raise RuntimeError(stderr.strip())

        shutil.rmtree(tmpdir)

    def cmake_generate(self):
        self.announce('running cmake generation', level=2)
        cmake_args = ['cmake', SOURCE_DIR] + self.generate_cmake_arguments()
        subprocess.check_call(
            cmake_args,
            cwd=self.build_temp_dir,
            env=os.environ.copy()
        )

    def cmake_install(self):
        self.announce('running cmake build', level=2)
        if self.is_windows() and not self.is_mingw():
            multi_proc = '/m'
        else:
            multi_proc = '-j{}'.format(multiprocessing.cpu_count())

        subprocess.check_call(
            [
                'cmake',
                '--build', '.',
                '--target', 'install',
                '--config', self.build_config,
                '--', multi_proc,
            ],
            cwd=self.build_temp_dir,
            env=os.environ.copy()
        )


# check the python version first
if (
    sys.version_info[0] < 2 or
    (sys.version_info[0] == 2 and sys.version_info[1] < 7)
):
    sys.exit(
        'OpenTimelineIO requires python2.7 or greater, detected version:'
        ' {}.{}'.format(
            sys.version_info[0],
            sys.version_info[1]
        )
    )


# Metadata that gets stamped into the __init__ files during the build phase.
PROJECT_METADATA = {
    "version": "0.15.0",
    "author": 'Contributors to the OpenTimelineIO project',
    "author_email": 'otio-discussion@lists.aswf.io',
    "license": 'Modified Apache 2.0 License',
}

METADATA_TEMPLATE = """
__version__ = "{version}"
__author__ = "{author}"
__author_email__ = "{author_email}"
__license__ = "{license}"
"""


def _append_version_info_to_init_scripts(build_lib):
    """Stamp PROJECT_METADATA into __init__ files."""

    for module, parentdir in [
            ("opentimelineio", "src/py-opentimelineio"),
            ("opentimelineio_contrib", "contrib"),
            ("opentimelineview", "src")
    ]:
        target_file = os.path.join(build_lib, module, "__init__.py")
        source_file = os.path.join(
            os.path.dirname(__file__),
            parentdir,
            module,
            "__init__.py"
        )

        # get the base data from the original file
        with open(source_file, 'r') as fi:
            src_data = fi.read()

        # write that + the suffix to the target file
        with open(target_file, 'w') as fo:
            fo.write(src_data)
            fo.write(METADATA_TEMPLATE.format(**PROJECT_METADATA))


class OTIO_build_py(setuptools.command.build_py.build_py):
    """Stamps PROJECT_METADATA into __init__ files."""

    def run(self):
        setuptools.command.build_py.build_py.run(self)

        if not self.dry_run:
            _append_version_info_to_init_scripts(self.build_lib)


def test_otio():
    """Discovers and runs tests"""
    try:
        # Clear the environment of a preset media linker
        del os.environ['OTIO_DEFAULT_MEDIA_LINKER']
    except KeyError:
        pass
    return unittest.TestLoader().discover('tests')


# copied from first paragraph of README.md
LONG_DESCRIPTION = """OpenTimelineIO is an interchange format and API for
editorial cut information. OTIO is not a container format for media, rather it
contains information about the order and length of cuts and references to
external media.

OTIO includes both a file format and an API for manipulating that format. It
also includes a plugin architecture for writing adapters to convert from/to
existing editorial timeline formats. It also implements a dependency- less
library for dealing strictly with time, opentime.

You can provide adapters for your video editing tool or pipeline as needed.
Each adapter allows for import/export between that proprietary tool and the
OpenTimelineIO format."""

setup(
    name='OpenTimelineIO',
    description='Editorial interchange format and API',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='http://opentimeline.io',
    project_urls={
        'Source':
            'https://github.com/AcademySoftwareFoundation/OpenTimelineIO',
        'Documentation':
            'https://opentimelineio.readthedocs.io/',
        'Issues':
            'https://github.com/AcademySoftwareFoundation/OpenTimelineIO/issues',
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Multimedia :: Video',
        'Topic :: Multimedia :: Video :: Display',
        'Topic :: Multimedia :: Video :: Non-Linear Editor',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: Other/Proprietary License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: OS Independent',
        'Natural Language :: English',
    ],

    keywords='film tv editing editorial edit non-linear edl time',

    platforms='any',

    package_data={
        'opentimelineio': [
            'adapters/builtin_adapters.plugin_manifest.json',
        ],
        'opentimelineio_contrib': [
            'adapters/contrib_adapters.plugin_manifest.json',
        ],
    },

    packages=(
        find_packages(where="src/py-opentimelineio") +  # opentimelineio
        find_packages(where="src") +  # opentimelineview
        find_packages(where="contrib", exclude=["opentimelineio_contrib.adapters.tests"])  # opentimelineio_contrib # noqa
    ),

    ext_modules=[
        CMakeExtension('_opentimelineio'),
        CMakeExtension('_opentime'),
    ],

    package_dir={
        'opentimelineio_contrib': 'contrib/opentimelineio_contrib',
        'opentimelineio': 'src/py-opentimelineio/opentimelineio',
        'opentimelineview': 'src/opentimelineview',
    },

    # Disallow 3.9.0 because of https://github.com/python/cpython/pull/22670
    python_requires='>2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*, !=3.9.0',  # noqa: E501

    install_requires=[
        'pyaaf2>=1.4,<1.7',
        'backports.tempfile; python_version<"3.0"',
        # Enables the builtins module in the XGES adapter
        'future; python_version<"3.0"',
        # Used in the otioz adapter to conform to unix paths
        'pathlib2; python_version<"3.0"'
    ],
    entry_points={
        'console_scripts': [
            'otiocat = opentimelineio.console.otiocat:main',
            'otioconvert = opentimelineio.console.otioconvert:main',
            'otiopluginfo = opentimelineio.console.otiopluginfo:main',
            'otiostat = opentimelineio.console.otiostat:main',
            'otiotool = opentimelineio.console.otiotool:main',
            'otioview = opentimelineview.console:main',
            (
                'otioautogen_serialized_schema_docs = '
                'opentimelineio.console.autogen_serialized_datamodel:main'
            ),
        ],
    },
    extras_require={
        'dev': [
            'check-manifest',
            'flake8>=3.5',
            'coverage>=4.5',
            'urllib3>=1.24.3'
        ],
        'view': [
            'PySide2~=5.11; platform.machine=="x86_64"',
            'PySide6~=6.2; platform.machine=="aarch64"'
        ]
    },

    test_suite='setup.test_otio',

    tests_require=[
        'mock;python_version<"3.3"',
    ],

    # because we need to open() the adapters manifest, we aren't zip-safe
    zip_safe=False,

    # The sequence of operations performed by setup.py is:
    #   OTIO_install::initialize
    #   OTIO_install::run
    #   OTIO_build_py::run
    #   the OpenTimelineIO egg is created
    #   OTIO_build_ext::initialize_options
    #   MANIFEST.in is read
    #   the lack of CHANGELOG.md is then reported.
    #   OTIO_build_ext::run is called
    #   OTIO_build_ext::build
    #   site-packages/opentimelineio* is populated with all scripts and extensions.
    #   pyc's are created for every python script.
    #   The egg is moved into site-packages,
    #   wrapper scripts for all the otiotools are created in a bin directory at the
    #   installation root.

    cmdclass={
        'build_py': OTIO_build_py,
        'build_ext': OTIO_build_ext,
    },

    # expand the project metadata dictionary to fill in those values
    **PROJECT_METADATA
)



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Unit tests for the rv reader plugin"""

# Please note that the test is geared towards linux install as I don't have
# a Windows or Mac available to me.

import os
import sys
import ast
import socket
import zipfile
import tempfile
import unittest
import shutil
import shlex
import time
import imp
import platform
from subprocess import call, Popen, PIPE

import opentimelineio as otio

RV_OTIO_READER_NAME = 'Example OTIO Reader'
RV_OTIO_READER_VERSION = '1.0'

RV_ROOT_DIR = os.getenv('OTIO_RV_ROOT_DIR', '')
RV_BIN_DIR = os.path.join(
    RV_ROOT_DIR,
    'MacOS' if platform.system() == 'Darwin' else 'bin'
)

RV_OTIO_READER_DIR = os.path.join(
    '..',
    'rv',
    'example_otio_reader'
)

# Import rvNetwork with imp to compensate for older RV's missing __init__.py
RV_NETWORK_MODULE = os.path.join(
    RV_ROOT_DIR,
    'src',
    'python',
    'network',
    'rvNetwork.py'
)
rvNetwork = imp.load_source('rvNetwork', RV_NETWORK_MODULE)

# Generate sample data
sample_timeline = otio.schema.Timeline(
    'my_timeline',
    global_start_time=otio.opentime.RationalTime(1, 24)
)
track = otio.schema.Track('v1')
bounds = [
    otio.schema.Box2d(
        otio.schema.V2d(0.0, 0.0),
        otio.schema.V2d(16.0, 9.0)
    ),  # sets viewing area
    otio.schema.Box2d(
        otio.schema.V2d(8.0, 0),
        otio.schema.V2d(24.0, 9.0)
    ),  # shifted right by half the viewing area
    otio.schema.Box2d(
        otio.schema.V2d(0.0, 0.0),
        otio.schema.V2d(8.0, 4.5)
    )  # scale to 1/4 of viewing area (lower left)
]

for clipnum, box in zip(range(1, 4), bounds):
    clip_name = 'clip{n}'.format(n=clipnum)
    track.append(
        otio.schema.Clip(
            clip_name,
            media_reference=otio.schema.ExternalReference(
                target_url="{clip_name}.mov".format(clip_name=clip_name),
                available_range=otio.opentime.TimeRange(
                    otio.opentime.RationalTime(1, 24),
                    otio.opentime.RationalTime(50, 24)
                ),
                available_image_bounds=box
            ),
            source_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(11, 24),
                otio.opentime.RationalTime(3, 24)
            )
        )
    )
sample_timeline.tracks.append(track)


@unittest.skipIf(
    "OTIO_RV_ROOT_DIR" not in os.environ,
    "OTIO_RV_ROOT_DIR not set."
)
@unittest.skipIf(
    (sys.version_info > (3, 0)),
    "RV OTIO reader plugin does not work in python 3."
)
class RVSessionAdapterReadTest(unittest.TestCase):
    def create_temp_dir(self):
        return tempfile.mkdtemp(prefix='rv_otio_reader')

    def test_create_rvpkg(self):
        temp_dir = self.create_temp_dir()
        package_path = create_rvpkg(temp_dir)

        self.assertTrue(os.path.exists(package_path))
        self.assertTrue(os.path.getsize(package_path) > 0)

        # Cleanup
        shutil.rmtree(temp_dir)

    def test_install_plugin(self):
        temp_dir = self.create_temp_dir()
        source_package_path = create_rvpkg(temp_dir)

        # Install package
        rc = install_package(source_package_path)

        # Check if install succeeded
        installed_package_path = os.path.join(
            temp_dir,
            'Packages',
            os.path.basename(source_package_path)
        )

        self.assertTrue(rc == 0)
        self.assertTrue(len(os.listdir(temp_dir)) > 1)
        self.assertTrue(os.path.exists(installed_package_path))

        # Make sure package is available in RV
        list_cmd = '{root}/rvpkg ' \
                   '-only {tmp_dir} -list'\
                   .format(
                       root=RV_BIN_DIR,
                       tmp_dir=temp_dir
                   )

        proc = Popen(shlex.split(list_cmd), stdout=PIPE)
        stdout, _ = proc.communicate()

        desired_result = \
            'I L - {version} "{package_name}" {pkg_path}'.format(
                version=RV_OTIO_READER_VERSION,
                package_name=RV_OTIO_READER_NAME,
                pkg_path=os.path.realpath(installed_package_path)
            )

        self.assertIn(desired_result, stdout.split('\n'))

        # Cleanup
        shutil.rmtree(temp_dir)

    def test_read_otio_file(self):
        # Install package
        temp_dir = self.create_temp_dir()
        source_package_path = create_rvpkg(temp_dir)
        install_package(source_package_path)

        env = os.environ.copy()
        env.update({'RV_SUPPORT_PATH': temp_dir})

        sample_file = tempfile.NamedTemporaryFile(
            'w',
            prefix='otio_data_',
            suffix='.otio',
            dir=temp_dir,
            delete=False
        )
        otio.adapters.write_to_file(sample_timeline, sample_file.name)
        run_cmd = '{root}/{exe} ' \
                  '-nc ' \
                  '-network ' \
                  '-networkHost localhost ' \
                  '-networkPort {port} ' \
                  '{sample_file}' \
                  .format(
                      exe='RV' if platform.system() == 'Darwin' else 'rv',
                      root=RV_BIN_DIR,
                      port=9876,
                      sample_file=sample_file.name
                  )
        proc = Popen(shlex.split(run_cmd), env=env)

        # Connect with RV
        rvc = rvNetwork.RvCommunicator()

        try:
            attempts = 0
            while not rvc.connected:
                attempts += 1
                rvc.connect('localhost', 9876)

                if not rvc.connected:
                    time.sleep(.5)

                if attempts == 20:
                    raise socket.error(
                        "Unable to connect to RV!"
                    )

            # some time can pass between the RV connection
            # and the complete startup of RV
            print("Waiting for RV startup to complete")
            time.sleep(10)

            # Check clips at positions
            clip1 = rv_media_name_at_frame(rvc, 1)
            self.assertEqual(clip1, 'clip1.mov')

            # note RV has a default res of 1280,720 when the media doesn't exist
            aspect_ratio = 1280.0 / 720.0

            clip1_scale, clip1_translate = rv_transform_at_frame(rvc, 1)
            self.assertEqual(clip1_scale, [1.0, 1.0])
            self.assertEqual(clip1_translate, [0.0, 0.0])

            clip2 = rv_media_name_at_frame(rvc, 4)
            self.assertEqual(clip2, 'clip2.mov')

            clip2_scale, clip2_translate = rv_transform_at_frame(rvc, 4)
            self.assertEqual(clip2_scale, [1.0, 1.0])

            self.assertAlmostEqual(clip2_translate[0], 0.5 * aspect_ratio)
            self.assertEqual(clip2_translate[1], 0)

            clip3 = rv_media_name_at_frame(rvc, 7)
            self.assertEqual(clip3, 'clip3.mov')

            clip3_scale, clip3_translate = rv_transform_at_frame(rvc, 7)
            self.assertEqual(clip3_scale, [0.5, 0.5])

            self.assertAlmostEqual(clip3_translate[0], -0.25 * aspect_ratio)
            self.assertEqual(clip3_translate[1], -0.25)

            rvc.disconnect()

        finally:
            # Cleanup
            proc.terminate()
            shutil.rmtree(temp_dir)


def create_rvpkg(temp_dir):
    package_path = os.path.join(
        temp_dir,
        'example_otio_reader_plugin-{version}.rvpkg'
        .format(version=RV_OTIO_READER_VERSION)
    )
    with zipfile.ZipFile(package_path, 'w') as pkg:
        for item in os.listdir(RV_OTIO_READER_DIR):
            pkg.write(os.path.join(RV_OTIO_READER_DIR, item), item)

    return package_path


def install_package(source_package_path):
    install_cmd = '{root}/rvpkg -force ' \
                  '-install ' \
                  '-add {tmp_dir} ' \
                  '{pkg_file}'.format(
                      root=RV_BIN_DIR,
                      tmp_dir=os.path.dirname(source_package_path),
                      pkg_file=source_package_path
                  )
    rc = call(shlex.split(install_cmd))
    return rc


def _exec_command(rvc, command, literal=True):
    response = rvc.sendEventAndReturn("remote-pyeval", command)
    return ast.literal_eval(response) if literal else response


def _source_at_frame(rvc, frame):
    return _exec_command(
        rvc,
        "rv.commands.sourcesAtFrame({0})".format(frame)
    )[0]


def rv_media_name_at_frame(rvc, frame):
    source_name = _source_at_frame(rvc, frame)
    return _exec_command(
        rvc,
        "rv.commands.sourceMedia('{0}')".format(source_name)
    )[0]


def rv_transform_at_frame(rvc, frame):
    source = _source_at_frame(rvc, frame)

    source_group = _exec_command(
        rvc,
        """rv.commands.nodeGroup('{0}')""".format(source),
        literal=False
    )

    transform = _exec_command(
        rvc,
        """rv.extra_commands.nodesInGroupOfType(
            '{0}', 'RVTransform2D')""".format(source_group)
    )[0]

    scale = _exec_command(
        rvc,
        """rv.commands.getFloatProperty(
            '{0}.transform.scale')""".format(transform)
    )

    translate = _exec_command(
        rvc,
        """rv.commands.getFloatProperty(
            '{0}.transform.translate')""".format(transform)
    )

    return scale, translate


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Test harness for Image Sequence References."""
import unittest
import sys

import opentimelineio as otio
import opentimelineio.test_utils as otio_test_utils


IS_PYTHON_2 = (sys.version_info < (3, 0))


class ImageSequenceReferenceTests(
    unittest.TestCase, otio_test_utils.OTIOAssertions
):
    def test_create(self):
        frame_policy = otio.schema.ImageSequenceReference.MissingFramePolicy.hold
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=5,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 30),
                otio.opentime.RationalTime(60, 30),
            ),
            frame_step=3,
            missing_frame_policy=frame_policy,
            rate=30,
            metadata={"custom": {"foo": "bar"}},
        )

        # Check Values
        self.assertEqual(ref.target_url_base, "file:///show/seq/shot/rndr/")
        self.assertEqual(ref.name_prefix, "show_shot.")
        self.assertEqual(ref.name_suffix, ".exr")
        self.assertEqual(ref.frame_zero_padding, 5)
        self.assertEqual(
            ref.available_range,
            otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 30),
                otio.opentime.RationalTime(60, 30),
            )
        )
        self.assertEqual(ref.frame_step, 3)
        self.assertEqual(ref.rate, 30)
        self.assertEqual(ref.metadata, {"custom": {"foo": "bar"}})
        self.assertEqual(
            ref.missing_frame_policy,
            otio.schema.ImageSequenceReference.MissingFramePolicy.hold,
        )

    @unittest.skipIf(IS_PYTHON_2, "unicode strings do funny things in python2")
    def test_str(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            start_frame=1,
            frame_step=3,
            rate=30,
            frame_zero_padding=5,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 30),
                otio.opentime.RationalTime(60, 30),
            ),
            metadata={"custom": {"foo": "bar"}},
            available_image_bounds=otio.schema.Box2d(
                otio.schema.V2d(0.0, 0.0),
                otio.schema.V2d(16.0, 9.0)
            ),
        )
        self.assertEqual(
            str(ref),
            'ImageSequenceReference('
            '"file:///show/seq/shot/rndr/", '
            '"show_shot.", '
            '".exr", '
            '1, '
            '3, '
            '30.0, '
            '5, '
            'MissingFramePolicy.error, '
            'TimeRange(RationalTime(0, 30), RationalTime(60, 30)), '
            'Box2d(V2d(0.0, 0.0), V2d(16.0, 9.0)), '
            "{'custom': {'foo': 'bar'}}"
            ')'
        )

    @unittest.skipIf(IS_PYTHON_2, "unicode strings do funny things in python2")
    def test_repr(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            start_frame=1,
            frame_step=3,
            rate=30,
            frame_zero_padding=5,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 30),
                otio.opentime.RationalTime(60, 30),
            ),
            available_image_bounds=otio.schema.Box2d(
                otio.schema.V2d(0.0, 0.0),
                otio.schema.V2d(16.0, 9.0)
            ),
            metadata={"custom": {"foo": "bar"}},
        )
        ref_value = (
            'ImageSequenceReference('
            "target_url_base='file:///show/seq/shot/rndr/', "
            "name_prefix='show_shot.', "
            "name_suffix='.exr', "
            'start_frame=1, '
            'frame_step=3, '
            'rate=30.0, '
            'frame_zero_padding=5, '
            'missing_frame_policy=<MissingFramePolicy.error: 0>, '
            'available_range={}, '
            'available_image_bounds=otio.schema.Box2d('
            'min=otio.schema.V2d(x=0.0, y=0.0), '
            'max=otio.schema.V2d(x=16.0, y=9.0)), '
            "metadata={{'custom': {{'foo': 'bar'}}}}"
            ')'.format(repr(ref.available_range))
        )
        self.assertEqual(repr(ref), ref_value)

    def test_serialize_roundtrip(self):
        frame_policy = otio.schema.ImageSequenceReference.MissingFramePolicy.hold
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=5,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 30),
                otio.opentime.RationalTime(60, 30),
            ),
            frame_step=3,
            rate=30,
            missing_frame_policy=frame_policy,
            metadata={"custom": {"foo": "bar"}},
        )

        encoded = otio.adapters.otio_json.write_to_string(ref)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(ref, decoded)

        encoded2 = otio.adapters.otio_json.write_to_string(decoded)
        self.assertEqual(encoded, encoded2)

        # Check Values
        self.assertEqual(
            decoded.target_url_base, "file:///show/seq/shot/rndr/"
        )
        self.assertEqual(decoded.name_prefix, "show_shot.")
        self.assertEqual(decoded.name_suffix, ".exr")
        self.assertEqual(decoded.frame_zero_padding, 5)
        self.assertEqual(
            decoded.available_range,
            otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 30),
                otio.opentime.RationalTime(60, 30),
            )
        )
        self.assertEqual(decoded.frame_step, 3)
        self.assertEqual(decoded.rate, 30)
        self.assertEqual(
            decoded.missing_frame_policy,
            otio.schema.ImageSequenceReference.MissingFramePolicy.hold
        )
        self.assertEqual(decoded.metadata, {"custom": {"foo": "bar"}})

    def test_deserialize_invalid_enum_value(self):
        encoded = """{
            "OTIO_SCHEMA": "ImageSequenceReference.1",
            "metadata": {
                "custom": {
                    "foo": "bar"
                }
            },
            "name": "",
            "available_range": {
                "OTIO_SCHEMA": "TimeRange.1",
                "duration": {
                    "OTIO_SCHEMA": "RationalTime.1",
                    "rate": 30.0,
                    "value": 60.0
                },
                "start_time": {
                    "OTIO_SCHEMA": "RationalTime.1",
                    "rate": 30.0,
                    "value": 0.0
                }
            },
            "target_url_base": "file:///show/seq/shot/rndr/",
            "name_prefix": "show_shot.",
            "name_suffix": ".exr",
            "start_frame": 1,
            "frame_step": 3,
            "rate": 30.0,
            "frame_zero_padding": 5,
            "missing_frame_policy": "BOGUS"
        }"""
        with self.assertRaises(ValueError):
            otio.adapters.otio_json.read_from_string(encoded)

    def test_number_of_images_in_sequence(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 24),
                otio.opentime.RationalTime(48, 24),
            ),
            rate=24,
        )

        self.assertEqual(ref.number_of_images_in_sequence(), 48)

    def test_number_of_images_in_sequence_with_skip(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 24),
                otio.opentime.RationalTime(48, 24),
            ),
            frame_step=2,
            rate=24,
        )

        self.assertEqual(ref.number_of_images_in_sequence(), 24)

        ref.frame_step = 3
        self.assertEqual(ref.number_of_images_in_sequence(), 16)

    def test_target_url_for_image_number(self):
        all_images_urls = [
            "file:///show/seq/shot/rndr/show_shot.{:04}.exr".format(i)
            for i in range(1, 49)
        ]
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 24),
                otio.opentime.RationalTime(48, 24),
            ),
            start_frame=1,
            frame_step=1,
            rate=24,
        )

        generated_urls = [
            ref.target_url_for_image_number(i)
            for i in range(ref.number_of_images_in_sequence())
        ]
        self.assertEqual(all_images_urls, generated_urls)

    def test_target_url_for_image_number_steps(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 24),
                otio.opentime.RationalTime(48, 24),
            ),
            start_frame=1,
            frame_step=2,
            rate=24,
        )

        all_images_urls = [
            "file:///show/seq/shot/rndr/show_shot.{:04}.exr".format(i)
            for i in range(1, 49, 2)
        ]
        generated_urls = [
            ref.target_url_for_image_number(i)
            for i in range(ref.number_of_images_in_sequence())
        ]
        self.assertEqual(all_images_urls, generated_urls)

        ref.frame_step = 3
        all_images_urls_threes = [
            "file:///show/seq/shot/rndr/show_shot.{:04}.exr".format(i)
            for i in range(1, 49, 3)
        ]
        generated_urls_threes = [
            ref.target_url_for_image_number(i)
            for i in range(ref.number_of_images_in_sequence())
        ]
        self.assertEqual(all_images_urls_threes, generated_urls_threes)

        ref.frame_step = 2
        ref.start_frame = 0
        all_images_urls_zero_first = [
            "file:///show/seq/shot/rndr/show_shot.{:04}.exr".format(i)
            for i in range(0, 48, 2)
        ]
        generated_urls_zero_first = [
            ref.target_url_for_image_number(i)
            for i in range(ref.number_of_images_in_sequence())
        ]
        self.assertEqual(all_images_urls_zero_first, generated_urls_zero_first)

    def test_target_url_for_image_number_with_missing_slash(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 24),
                otio.opentime.RationalTime(48, 24),
            ),
            start_frame=1,
            frame_step=1,
            rate=24,
        )

        self.assertEqual(
            ref.target_url_for_image_number(0),
            "file:///show/seq/shot/rndr/show_shot.0001.exr"
        )

    def test_abstract_target_url(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 24),
                otio.opentime.RationalTime(48, 24),
            ),
            start_frame=1,
            frame_step=1,
            rate=24,
        )

        self.assertEqual(
            ref.abstract_target_url("@@@@"),
            "file:///show/seq/shot/rndr/show_shot.@@@@.exr"
        )

    def test_abstract_target_url_with_missing_slash(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 24),
                otio.opentime.RationalTime(48, 24),
            ),
            start_frame=1,
            frame_step=1,
            rate=24,
        )

        self.assertEqual(
            ref.abstract_target_url("@@@@"),
            "file:///show/seq/shot/rndr/show_shot.@@@@.exr"
        )

    def test_presentation_time_for_image_number(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(0, 24),
                otio.opentime.RationalTime(48, 24),
            ),
            start_frame=1,
            frame_step=2,
            rate=24,
        )

        reference_values = [
            otio.opentime.RationalTime(i * 2, 24) for i in range(24)
        ]

        generated_values = [
            ref.presentation_time_for_image_number(i)
            for i in range(ref.number_of_images_in_sequence())
        ]

        self.assertEqual(generated_values, reference_values)

    def test_presentation_time_for_image_number_with_offset(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(12, 24),
                otio.opentime.RationalTime(48, 24),
            ),
            start_frame=1,
            frame_step=2,
            rate=24,
        )

        first_frame_time = otio.opentime.RationalTime(12, 24)
        reference_values = [
            first_frame_time + otio.opentime.RationalTime(i * 2, 24)
            for i in range(24)
        ]

        generated_values = [
            ref.presentation_time_for_image_number(i)
            for i in range(ref.number_of_images_in_sequence())
        ]

        self.assertEqual(generated_values, reference_values)

    def test_end_frame(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(12, 24),
                otio.opentime.RationalTime(48, 24),
            ),
            start_frame=1,
            frame_step=1,
            rate=24,
        )

        self.assertEqual(ref.end_frame(), 48)

        # Frame step should not affect this
        ref.frame_step = 2
        self.assertEqual(ref.end_frame(), 48)

    def test_end_frame_with_offset(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(12, 24),
                otio.opentime.RationalTime(48, 24),
            ),
            start_frame=101,
            frame_step=1,
            rate=24,
        )

        self.assertEqual(ref.end_frame(), 148)

        # Frame step should not affect this
        ref.frame_step = 2
        self.assertEqual(ref.end_frame(), 148)

    def test_frame_for_time(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(12, 24),
                otio.opentime.RationalTime(48, 24),
            ),
            start_frame=1,
            frame_step=1,
            rate=24,
        )

        # The start time should be frame 1
        self.assertEqual(
            ref.frame_for_time(ref.available_range.start_time), 1
        )

        # Test a sample in the middle
        self.assertEqual(
            ref.frame_for_time(otio.opentime.RationalTime(15, 24)), 4
        )

        # The end time (inclusive) should map to the last frame number
        self.assertEqual(
            ref.frame_for_time(ref.available_range.end_time_inclusive()), 48
        )

        # make sure frame step and RationalTime rate have no effect
        ref.frame_step = 2
        self.assertEqual(
            ref.frame_for_time(otio.opentime.RationalTime(118, 48)), 48
        )

    def test_frame_for_time_out_of_range(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(12, 30),
                otio.opentime.RationalTime(60, 30),
            ),
            start_frame=1,
            frame_step=1,
            rate=30,
        )
        with self.assertRaises(ValueError):
            ref.frame_for_time(otio.opentime.RationalTime(73, 30))

    def test_frame_range_for_time_range(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(12, 24),
                otio.opentime.RationalTime(60, 24),
            ),
            start_frame=1,
            frame_step=1,
            rate=24,
        )
        time_range = otio.opentime.TimeRange(
            otio.opentime.RationalTime(24, 24),
            otio.opentime.RationalTime(17, 24),
        )

        self.assertEqual(ref.frame_range_for_time_range(time_range), (13, 29))

    def test_frame_range_for_time_range_out_of_available_image_bounds(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(12, 24),
                otio.opentime.RationalTime(60, 24),
            ),
            start_frame=1,
            frame_step=1,
            rate=24,
        )
        time_range = otio.opentime.TimeRange(
            otio.opentime.RationalTime(24, 24),
            otio.opentime.RationalTime(60, 24),
        )

        with self.assertRaises(ValueError):
            ref.frame_range_for_time_range(time_range)

    def test_negative_frame_numbers(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.RationalTime(12, 24),
                otio.opentime.RationalTime(48, 24),
            ),
            start_frame=-1,
            frame_step=2,
            rate=24,
        )

        self.assertEqual(ref.number_of_images_in_sequence(), 24)
        self.assertEqual(
            ref.presentation_time_for_image_number(0),
            otio.opentime.RationalTime(12, 24),
        )
        self.assertEqual(
            ref.presentation_time_for_image_number(1),
            otio.opentime.RationalTime(14, 24),
        )
        self.assertEqual(
            ref.presentation_time_for_image_number(2),
            otio.opentime.RationalTime(16, 24),
        )
        self.assertEqual(
            ref.presentation_time_for_image_number(23),
            otio.opentime.RationalTime(58, 24),
        )

        self.assertEqual(
            ref.target_url_for_image_number(0),
            "file:///show/seq/shot/rndr/show_shot.-0001.exr",
        )

        self.assertEqual(
            ref.target_url_for_image_number(1),
            "file:///show/seq/shot/rndr/show_shot.0001.exr",
        )
        self.assertEqual(
            ref.target_url_for_image_number(2),
            "file:///show/seq/shot/rndr/show_shot.0003.exr",
        )
        self.assertEqual(
            ref.target_url_for_image_number(17),
            "file:///show/seq/shot/rndr/show_shot.0033.exr",
        )
        self.assertEqual(
            ref.target_url_for_image_number(23),
            "file:///show/seq/shot/rndr/show_shot.0045.exr",
        )

        # Check values by ones
        ref.frame_step = 1
        for i in range(1, ref.number_of_images_in_sequence()):
            self.assertEqual(
                ref.target_url_for_image_number(i),
                "file:///show/seq/shot/rndr/show_shot.{:04}.exr".format(i - 1),
            )

    def test_target_url_for_image_number_with_missing_timing_info(self):
        ref = otio.schema.ImageSequenceReference(
            "file:///show/seq/shot/rndr/",
            "show_shot.",
            ".exr",
            frame_zero_padding=4,
            start_frame=1,
            frame_step=1,
            rate=24,
        )

        # Make sure the right error and a useful message raised when
        # source_range is either un-set or zero duration.
        with self.assertRaises(IndexError) as exception_manager:
            ref.target_url_for_image_number(0)

        self.assertEqual(
            str(exception_manager.exception),
            "Zero duration sequences has no frames.",
        )

        ref.available_range = otio.opentime.TimeRange(
            otio.opentime.RationalTime(12, 24),
            otio.opentime.RationalTime(0, 1),
        )

        with self.assertRaises(IndexError) as exception_manager:
            ref.target_url_for_image_number(0)

        self.assertEqual(
            str(exception_manager.exception),
            "Zero duration sequences has no frames.",
        )

        # Set the duration and make sure a similarly useful message comes
        # when rate is un-set.
        ref.available_range = otio.opentime.TimeRange(
            otio.opentime.RationalTime(12, 24),
            otio.opentime.RationalTime(48, 24),
        )
        ref.rate = 0

        with self.assertRaises(IndexError) as exception_manager:
            ref.target_url_for_image_number(0)

        self.assertEqual(
            str(exception_manager.exception),
            "Zero rate sequence has no frames.",
        )

    def test_clone(self):
        """ ensure that deeopcopy/clone function """
        isr = otio.schema.ImageSequenceReference()

        try:
            import copy
            cln = copy.deepcopy(isr)
            cln = isr.clone()
        except ValueError as exc:
            self.fail("Cloning raised an exception: {}".format(exc))

        self.assertJsonEqual(isr, cln)

    def test_target_url_for_image_number_with_blank_target_url_base(self):
        ref = otio.schema.ImageSequenceReference(
            name_prefix="myfilename.",
            name_suffix=".exr",
            start_frame=101,
            rate=24,
            frame_zero_padding=4,
            available_range=otio.opentime.TimeRange(
                otio.opentime.from_timecode("01:25:30:04", rate=24),
                duration=otio.opentime.from_frames(48, 24)
            ),
        )

        self.assertEqual(
            ref.target_url_for_image_number(0), "myfilename.0101.exr"
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Test file for the track algorithms library."""

import unittest

import opentimelineio as otio
import opentimelineio.test_utils as otio_test_utils


class TimelineTrimmingTests(unittest.TestCase, otio_test_utils.OTIOAssertions):
    """ test harness for timeline trimming function """

    def make_sample_timeline(self):
        result = otio.adapters.read_from_string(
            """
            {
                "OTIO_SCHEMA": "Timeline.1",
                "metadata": {},
                "name": null,
                "tracks": {
                    "OTIO_SCHEMA": "Stack.1",
                    "children": [
                        {
                            "OTIO_SCHEMA": "Track.1",
                            "children": [
                                {
                                    "OTIO_SCHEMA": "Clip.1",
                                    "effects": [],
                                    "markers": [],
                                    "media_reference": null,
                                    "metadata": {},
                                    "name": "A",
                                    "source_range": {
                                        "OTIO_SCHEMA": "TimeRange.1",
                                        "duration": {
                                            "OTIO_SCHEMA": "RationalTime.1",
                                            "rate": 24,
                                            "value": 50
                                        },
                                        "start_time": {
                                            "OTIO_SCHEMA": "RationalTime.1",
                                            "rate": 24,
                                            "value": 0.0
                                        }
                                    }
                                },
                                {
                                    "OTIO_SCHEMA": "Clip.1",
                                    "effects": [],
                                    "markers": [],
                                    "media_reference": null,
                                    "metadata": {},
                                    "name": "B",
                                    "source_range": {
                                        "OTIO_SCHEMA": "TimeRange.1",
                                        "duration": {
                                            "OTIO_SCHEMA": "RationalTime.1",
                                            "rate": 24,
                                            "value": 50
                                        },
                                        "start_time": {
                                            "OTIO_SCHEMA": "RationalTime.1",
                                            "rate": 24,
                                            "value": 0.0
                                        }
                                    }
                                },
                                {
                                    "OTIO_SCHEMA": "Clip.1",
                                    "effects": [],
                                    "markers": [],
                                    "media_reference": null,
                                    "metadata": {},
                                    "name": "C",
                                    "source_range": {
                                        "OTIO_SCHEMA": "TimeRange.1",
                                        "duration": {
                                            "OTIO_SCHEMA": "RationalTime.1",
                                            "rate": 24,
                                            "value": 50
                                        },
                                        "start_time": {
                                            "OTIO_SCHEMA": "RationalTime.1",
                                            "rate": 24,
                                            "value": 0.0
                                        }
                                    }
                                }
                            ],
                            "effects": [],
                            "kind": "Video",
                            "markers": [],
                            "metadata": {},
                            "name": "Sequence1",
                            "source_range": null
                        }
                    ],
                    "effects": [],
                    "markers": [],
                    "metadata": {},
                    "name": "tracks",
                    "source_range": null
                }
            }""",
            "otio_json"
        )

        return result, result.tracks[0]

    def test_trim_to_existing_range(self):
        original_timeline, original_track = self.make_sample_timeline()
        self.assertEqual(
            original_track.trimmed_range(),
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(150, 24)
            )
        )

        # trim to the exact range it already has
        trimmed = otio.algorithms.timeline_trimmed_to_range(
            original_timeline,
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(150, 24)
            )
        )
        # it shouldn't have changed at all
        self.assertIsOTIOEquivalentTo(original_timeline, trimmed)

    def test_trim_to_longer_range(self):
        original_timeline, original_track = self.make_sample_timeline()
        # trim to a larger range
        trimmed = otio.algorithms.timeline_trimmed_to_range(
            original_timeline,
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(-10, 24),
                duration=otio.opentime.RationalTime(160, 24)
            )
        )
        # it shouldn't have changed at all
        self.assertJsonEqual(original_timeline, trimmed)

    def test_trim_front(self):
        original_timeline, original_track = self.make_sample_timeline()
        # trim off the front (clip A and part of B)
        trimmed = otio.algorithms.timeline_trimmed_to_range(
            original_timeline,
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(60, 24),
                duration=otio.opentime.RationalTime(90, 24)
            )
        )
        self.assertNotEqual(original_timeline, trimmed)
        trimmed = trimmed.tracks[0]

        self.assertEqual(len(trimmed), 2)
        self.assertEqual(
            trimmed.trimmed_range(),
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(90, 24)
            )
        )
        # did clip B get trimmed?
        self.assertEqual(trimmed[0].name, "B")
        self.assertEqual(
            trimmed[0].trimmed_range(),
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(10, 24),
                duration=otio.opentime.RationalTime(40, 24)
            )
        )
        # clip C should have been left alone
        self.assertIsOTIOEquivalentTo(trimmed[1], original_track[2])

    def test_trim_end(self):
        original_timeline, original_track = self.make_sample_timeline()
        # trim off the end (clip C and part of B)
        trimmed_timeline = otio.algorithms.timeline_trimmed_to_range(
            original_timeline,
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(90, 24)
            )
        )
        # rest of the tests are on the track
        trimmed = trimmed_timeline.tracks[0]

        self.assertNotEqual(original_timeline, trimmed)
        self.assertEqual(len(trimmed), 2)
        self.assertEqual(
            trimmed.trimmed_range(),
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(90, 24)
            )
        )

        # clip A should have been left alone
        self.assertIsOTIOEquivalentTo(trimmed[0], original_track[0])
        # did clip B get trimmed?
        self.assertEqual(trimmed[1].name, "B")
        self.assertEqual(
            trimmed[1].trimmed_range(),
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(0, 24),
                duration=otio.opentime.RationalTime(40, 24)
            )
        )

    def test_trim_with_transitions(self):
        original_timeline, original_track = self.make_sample_timeline()
        self.assertEqual(
            otio.opentime.RationalTime(150, 24),
            original_timeline.duration()
        )
        self.assertEqual(len(original_track), 3)

        # add a transition
        tr = otio.schema.Transition(
            in_offset=otio.opentime.RationalTime(12, 24),
            out_offset=otio.opentime.RationalTime(20, 24)
        )
        original_track.insert(1, tr)
        self.assertEqual(len(original_track), 4)
        self.assertEqual(
            otio.opentime.RationalTime(150, 24),
            original_timeline.duration()
        )

        # if you try to sever a Transition in the middle it should fail
        with self.assertRaises(otio.exceptions.CannotTrimTransitionsError):
            trimmed = otio.algorithms.timeline_trimmed_to_range(
                original_timeline,
                otio.opentime.TimeRange(
                    start_time=otio.opentime.RationalTime(5, 24),
                    duration=otio.opentime.RationalTime(50, 24)
                )
            )

        with self.assertRaises(otio.exceptions.CannotTrimTransitionsError):
            trimmed = otio.algorithms.timeline_trimmed_to_range(
                original_timeline,
                otio.opentime.TimeRange(
                    start_time=otio.opentime.RationalTime(45, 24),
                    duration=otio.opentime.RationalTime(50, 24)
                )
            )

        trimmed = otio.algorithms.timeline_trimmed_to_range(
            original_timeline,
            otio.opentime.TimeRange(
                start_time=otio.opentime.RationalTime(25, 24),
                duration=otio.opentime.RationalTime(50, 24)
            )
        )
        self.assertNotEqual(original_timeline, trimmed)

        expected = otio.adapters.read_from_string(
            """
            {
                "OTIO_SCHEMA": "Timeline.1",
                "metadata": {},
                "name": null,
                "tracks": {
                    "OTIO_SCHEMA": "Stack.1",
                    "children": [
                        {
                            "OTIO_SCHEMA": "Track.1",
                            "children": [
                                {
                                    "OTIO_SCHEMA": "Clip.1",
                                    "effects": [],
                                    "markers": [],
                                    "media_reference": null,
                                    "metadata": {},
                                    "name": "A",
                                    "source_range": {
                                        "OTIO_SCHEMA": "TimeRange.1",
                                        "duration": {
                                            "OTIO_SCHEMA": "RationalTime.1",
                                            "rate": 24,
                                            "value": 25
                                        },
                                        "start_time": {
                                            "OTIO_SCHEMA": "RationalTime.1",
                                            "rate": 24,
                                            "value": 25.0
                                        }
                                    }
                                },
                                {
                                    "OTIO_SCHEMA": "Transition.1",
                                    "in_offset": {
                                        "OTIO_SCHEMA": "RationalTime.1",
                                        "rate": 24,
                                        "value": 12
                                    },
                                    "metadata": {},
                                    "name": null,
                                    "out_offset": {
                                        "OTIO_SCHEMA": "RationalTime.1",
                                        "rate": 24,
                                        "value": 20
                                    },
                                    "transition_type": null
                                },
                                {
                                    "OTIO_SCHEMA": "Clip.1",
                                    "effects": [],
                                    "markers": [],
                                    "media_reference": null,
                                    "metadata": {},
                                    "name": "B",
                                    "source_range": {
                                        "OTIO_SCHEMA": "TimeRange.1",
                                        "duration": {
                                            "OTIO_SCHEMA": "RationalTime.1",
                                            "rate": 24,
                                            "value": 25
                                        },
                                        "start_time": {
                                            "OTIO_SCHEMA": "RationalTime.1",
                                            "rate": 24,
                                            "value": 0.0
                                        }
                                    }
                                }
                            ],
                            "effects": [],
                            "kind": "Video",
                            "markers": [],
                            "metadata": {},
                            "name": "Sequence1",
                            "source_range": null
                        }
                    ],
                    "effects": [],
                    "markers": [],
                    "metadata": {},
                    "name": "tracks",
                    "source_range": null
                }
            }
            """,
            "otio_json"
        )

        self.assertJsonEqual(expected, trimmed)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

import os
import unittest

import opentimelineio as otio


# handle python2 vs python3 difference
try:
    from tempfile import TemporaryDirectory  # noqa: F401
    import tempfile
except ImportError:
    # XXX: python2.7 only
    from backports import tempfile

# Reference data
SAMPLE_DATA_DIR = os.path.join(os.path.dirname(__file__), "sample_data")
HLS_EXAMPLE_PATH = os.path.join(SAMPLE_DATA_DIR, "v1_prog_index.m3u8")

# Load the adapter module using otio
hls_playlist = otio.adapters.from_name("hls_playlist").module()

MEM_PLAYLIST_REF_VALUE = """#EXTM3U
#EXT-X-VERSION:7
#EXT-X-INDEPENDENT-SEGMENTS
#EXT-X-MEDIA-SEQUENCE:1
#EXT-X-PLAYLIST-TYPE:VOD
#EXT-X-TARGETDURATION:2
#EXT-X-MAP:BYTERANGE="729@0",URI="media-video-1.mp4"
#EXTINF:2.00200,
#EXT-X-BYTERANGE:534220@1361
video1.mp4
#EXT-X-ENDLIST"""

MEM_MASTER_PLAYLIST_REF_VALUE = """#EXTM3U
#EXT-X-VERSION:6
#EXT-X-MEDIA:GROUP-ID="aud1",NAME="a1",TYPE=AUDIO,URI="a1/prog_index.m3u8"
#EXT-X-STREAM-INF:AUDIO="aud1",BANDWIDTH=135801,CODECS="avc.test,aac.test",FRAME-RATE=23.976,RESOLUTION=1920x1080
v1/prog_index.m3u8"""

MEM_IFRAME_MASTER_PLAYLIST_REF_VALUE = """#EXTM3U
#EXT-X-VERSION:6
#EXT-X-MEDIA:GROUP-ID="aud1",NAME="a1",TYPE=AUDIO,URI="a1/prog_index.m3u8"
#EXT-X-I-FRAME-STREAM-INF:BANDWIDTH=123456,CODECS="avc.test",RESOLUTION=1920x1080,URI="v1/iframe_index.m3u8"
#EXT-X-STREAM-INF:AUDIO="aud1",BANDWIDTH=135801,CODECS="avc.test,aac.test",FRAME-RATE=23.976,RESOLUTION=1920x1080
v1/prog_index.m3u8"""

MEM_COMPLEX_MASTER_PLAYLIST_REF_VALUE = """#EXTM3U
#EXT-X-VERSION:6
#EXT-X-MEDIA:GROUP-ID="aud1",NAME="a1",TYPE=AUDIO,URI="a1/prog_index.m3u8"
#EXT-X-I-FRAME-STREAM-INF:BANDWIDTH=123456,CODECS="avc.test",RESOLUTION=1920x1080,URI="v1/iframe_index.m3u8"
#EXT-X-I-FRAME-STREAM-INF:BANDWIDTH=12345,CODECS="avc.test",RESOLUTION=720x480,URI="v2/iframe_index.m3u8"
#EXT-X-STREAM-INF:AUDIO="aud1",BANDWIDTH=135801,CODECS="avc.test,aac.test",FRAME-RATE=23.976,RESOLUTION=1920x1080
v1/prog_index.m3u8
#EXT-X-STREAM-INF:AUDIO="aud1",BANDWIDTH=24690,CODECS="avc.test,aac.test",FRAME-RATE=23.976,RESOLUTION=720x480
v2/prog_index.m3u8"""

MEM_SINGLE_TRACK_MASTER_PLAYLIST_REF_VALUE = """#EXTM3U
#EXT-X-VERSION:6
#EXT-X-I-FRAME-STREAM-INF:BANDWIDTH=123456,CODECS="avc.test",RESOLUTION=1920x1080,URI="v1/iframe_index.m3u8"
#EXT-X-STREAM-INF:BANDWIDTH=123456,CODECS="avc.test",FRAME-RATE=23.976,RESOLUTION=1920x1080
v1/prog_index.m3u8"""


class HLSPlaylistDataStructuresTest(unittest.TestCase):
    """ Test the lower-level HLS Data structures """

    def test_hls_attribute_list(self):
        """ Test the HLS adapter's attribute list parser """
        attribute_list_string = 'HEXTHING=0xFAF,FIRST-NEGFLOAT=-1.25,'\
            'STRTHING="foo, bar",DECIMALTHING=123456,FLOATTHING=1.233,'\
            'ENUMTHING=0xeS1,NEGFLOAT=-3.14'
        attribute_list_dictionary = {
            "HEXTHING": 0xFAF,
            "FIRST-NEGFLOAT": -1.25,
            "STRTHING": "foo, bar",
            "DECIMALTHING": 123456,
            "FLOATTHING": 1.233,
            "ENUMTHING": "0xeS1",
            "NEGFLOAT": -3.14
        }
        attr_list = hls_playlist.AttributeList.from_string(
            attribute_list_string
        )

        self.assertEqual(len(attr_list), len(attribute_list_dictionary))
        for attrName, attrValue in attr_list.items():
            self.assertEqual(attrValue, attribute_list_dictionary[attrName])

    def test_playlist_tag_exclusivity(self):
        """ Test that mutually-exclusive tag types don't overlap """
        # see sections 4.3.2, 4.3.3, and 4.3.4 of
        # draft-pantos-http-live-streaming for more information about these
        # constraints

        non_master_tags = hls_playlist.MEDIA_SEGMENT_TAGS.union(
            hls_playlist.MEDIA_PLAYLIST_TAGS)

        common_tags = non_master_tags.intersection(
            hls_playlist.MASTER_PLAYLIST_TAGS)
        self.assertEqual(len(common_tags), 0)


class HLSPMedialaylistAdapterTest(unittest.TestCase):
    """ Test the HLS Playlist adapter media playlist functionality """

    def test_media_pl_from_mem(self):
        t = otio.schema.Timeline()
        track = otio.schema.Track("v1")
        track.metadata['HLS'] = {
            "EXT-X-INDEPENDENT-SEGMENTS": None,
            "EXT-X-PLAYLIST-TYPE": "VOD"
        }
        t.tracks.append(track)

        # Make a prototype media ref with the segment's initialization metadata
        segmented_media_ref = otio.schema.ExternalReference(
            target_url='video1.mp4',
            metadata={
                "streaming": {
                    "init_byterange": {
                        "byte_count": 729,
                        "byte_offset": 0
                    },
                    "init_uri": "media-video-1.mp4"
                }
            }
        )

        # Make a copy of the media ref specifying the byte range for the
        # segment
        media_ref1 = segmented_media_ref.deepcopy()
        media_ref1.available_range = otio.opentime.TimeRange(
            otio.opentime.RationalTime(0, 1),
            otio.opentime.RationalTime(2.002, 1)
        )
        media_ref1.metadata['streaming'].update(
            {
                "byte_count": 534220,
                "byte_offset": 1361
            }
        )

        # make the segment and append it
        segment1 = otio.schema.Clip(media_reference=media_ref1)
        track.append(segment1)

        # Write out and validate the playlist
        with tempfile.TemporaryDirectory() as temp_dir:
            media_pl_tmp_path = os.path.join(
                temp_dir,
                "test_media_pl_from_mem.m3u8"
            )
            otio.adapters.write_to_file(t, media_pl_tmp_path)

            with open(media_pl_tmp_path) as f:
                pl_string = f.read()

        # Compare against the reference value
        self.assertEqual(pl_string, MEM_PLAYLIST_REF_VALUE)

    def _validate_sample_playlist(self, timeline):
        # Validate the track count
        self.assertEqual(len(timeline.tracks), 1)
        track = timeline.tracks[0]

        # Validate the track global metadata
        self.assertNotEqual(track.metadata['HLS'], {})
        track_metadata = track.metadata['HLS']
        self.assertEqual(track_metadata, {
            'EXT-X-INDEPENDENT-SEGMENTS': None,
            'EXT-X-VERSION': '7',
            'EXT-X-PLAYLIST-TYPE': 'VOD'}
        )

        # There are 50 segments (clips)
        # Validate the count, "sequence_num", and durations
        self.assertEqual(len(track), 50)
        start_seq_num = int(track[0].metadata['streaming']['sequence_num'])
        segment_durations = otio.opentime.RationalTime(1.001, 1)
        for seq_num, clip in enumerate(track, start_seq_num):
            self.assertEqual(
                clip.metadata['streaming']['sequence_num'],
                seq_num
            )
            if seq_num < 50:
                self.assertEqual(clip.duration(), segment_durations)
            else:
                # The last segment has a shorter duration
                self.assertEqual(
                    clip.duration(),
                    otio.opentime.RationalTime(0.83417, 1)
                )

        # Spot-check a segment
        segment_5 = track[4]
        seg_5_media_ref = segment_5.media_reference
        seg_5_ref_streaming_metadata = seg_5_media_ref.metadata['streaming']
        self.assertEqual(
            seg_5_ref_streaming_metadata['byte_count'],
            593718
        )
        self.assertEqual(
            seg_5_ref_streaming_metadata['byte_offset'],
            2430668
        )
        self.assertEqual(
            seg_5_ref_streaming_metadata['init_byterange']['byte_count'],
            729
        )
        self.assertEqual(
            seg_5_ref_streaming_metadata['init_byterange']['byte_offset'],
            0
        )
        self.assertEqual(
            seg_5_ref_streaming_metadata['init_uri'],
            "media-video-1.mp4"
        )
        self.assertEqual(
            seg_5_media_ref.target_url,
            "media-video-1.mp4"
        )

    def test_media_roundtrip(self):
        hls_path = HLS_EXAMPLE_PATH
        timeline = otio.adapters.read_from_file(hls_path)

        # validate the read-in playlist matches reference data
        self._validate_sample_playlist(timeline)

        # Write out and validate both playlists have the same lines
        with tempfile.TemporaryDirectory() as temp_dir:
            media_pl_tmp_path = os.path.join(
                temp_dir,
                "test_media_roundtrip.m3u8"
            )
            otio.adapters.write_to_file(timeline, media_pl_tmp_path)

            # Read in both playlists
            with open(hls_path) as f:
                reference_lines = f.readlines()

            with open(media_pl_tmp_path) as f:
                adapter_out_lines = f.readlines()

            # Using otio as well
            in_timeline = otio.adapters.read_from_file(media_pl_tmp_path)

        # Strip newline chars
        reference_lines = [line.strip('\n') for line in reference_lines]
        adapter_out_lines = [line.strip('\n') for line in adapter_out_lines]

        # Compare the lines
        self.assertEqual(reference_lines, adapter_out_lines)

        # validate the otio of the playlist we wrote
        self._validate_sample_playlist(in_timeline)

    def test_media_segment_size(self):
        hls_path = HLS_EXAMPLE_PATH
        timeline = otio.adapters.read_from_file(hls_path)

        # validate the read-in playlist matches reference data
        self._validate_sample_playlist(timeline)

        # Set the sement size to ~six seconds
        timeline_streaming_md = timeline.metadata.setdefault('streaming', {})
        seg_min_duration = otio.opentime.RationalTime(6, 1)
        timeline_streaming_md['min_segment_duration'] = seg_min_duration
        seg_max_duration = otio.opentime.RationalTime(
            (60 * 60 * 24),
            1
        )
        timeline_streaming_md['max_segment_duration'] = seg_max_duration

        # Write out the playlist
        with tempfile.TemporaryDirectory() as temp_dir:
            media_pl_tmp_path = os.path.join(
                temp_dir,
                "test_media_segment_size.m3u8"
            )
            otio.adapters.write_to_file(timeline, media_pl_tmp_path)

            # Read in the playlist
            in_timeline = otio.adapters.read_from_file(media_pl_tmp_path)

        # Pick a duration that segments won't exceed but is less than max
        seg_upper_duration = otio.opentime.RationalTime(7, 1)

        # When reading an HLS playlist, segments become clips. Check clip
        # durations (except the last one since it's the leftover)
        for clip in in_timeline.tracks[0][:-1]:
            self.assertTrue(clip.duration() >= seg_min_duration)
            self.assertTrue(clip.duration() < seg_upper_duration)

        # Check the last segment duration
        last_clip = in_timeline.tracks[0][-1]
        self.assertTrue(last_clip.duration() < seg_min_duration)
        self.assertTrue(
            last_clip.duration() > otio.opentime.RationalTime(
                0, 1
            )
        )

    def test_iframe_segment_size(self):
        hls_path = HLS_EXAMPLE_PATH
        timeline = otio.adapters.read_from_file(hls_path)

        # the reference playlist is one segment per keyframe, pluck the first
        # segment duration as reference for keyframe duration
        keyframe_duration = timeline.tracks[0][0].duration()

        # validate the read-in playlist matches reference data
        self._validate_sample_playlist(timeline)

        # Set the sement size to ~six seconds
        timeline_streaming_md = timeline.metadata.setdefault('streaming', {})
        seg_min_duration = otio.opentime.RationalTime(6, 1)
        timeline_streaming_md['min_segment_duration'] = seg_min_duration
        seg_max_duration = otio.opentime.RationalTime(
            (60 * 60 * 24),
            1
        )
        timeline_streaming_md['max_segment_duration'] = seg_max_duration

        # Configure the playlist to be an iframe list
        track_hls_metadata = timeline.tracks[0].metadata['HLS']
        del track_hls_metadata['EXT-X-INDEPENDENT-SEGMENTS']
        track_hls_metadata['EXT-X-I-FRAMES-ONLY'] = None

        # Write out the playlist
        with tempfile.TemporaryDirectory() as temp_dir:
            media_pl_tmp_path = os.path.join(
                temp_dir,
                "test_iframe_segment_size.m3u8"
            )
            otio.adapters.write_to_file(timeline, media_pl_tmp_path)

            # Read in the playlist
            in_timeline = otio.adapters.read_from_file(media_pl_tmp_path)
            with open(media_pl_tmp_path) as f:
                pl_lines = f.readlines()
            pl_lines = [line.strip('\n') for line in pl_lines]

        # validate the TARGETDURATION value is correct
        self.assertTrue('#EXT-X-TARGETDURATION:6' in pl_lines)
        self.assertTrue('#EXT-X-MEDIA-SEQUENCE:0' in pl_lines)
        self.assertEqual(len(timeline.tracks), len(in_timeline.tracks))
        self.assertEqual(len(timeline.tracks[0]), len(in_timeline.tracks[0]))

        # The segments should all be 1.001 seconds like the original input
        seg_upper_duration = otio.opentime.RationalTime(1.1, 1)

        # When reading an HLS playlist, segments become clips. Check clip
        # durations (except the last one since it's the leftover)
        for clip in in_timeline.tracks[0][:-1]:
            self.assertTrue(clip.duration() == keyframe_duration)
            self.assertTrue(clip.duration() < seg_upper_duration)

        # Check the last segment duration
        last_clip = in_timeline.tracks[0][-1]
        self.assertTrue(last_clip.duration() < seg_min_duration)
        self.assertTrue(
            last_clip.duration() > otio.opentime.RationalTime(0, 1)
        )


class HLSPMasterPlaylistAdapterTest(unittest.TestCase):
    """ Test the HLS Playlist adapter master playlist functionality """

    def test_simple_master_pl_from_mem(self):
        t = otio.schema.Timeline()

        # add a video track
        vtrack = otio.schema.Track(
            "v1",
            kind=otio.schema.TrackKind.Video
        )
        vtrack.metadata.update(
            {
                'streaming': {
                    'bandwidth': 123456,
                    'codec': 'avc.test',
                    'width': 1920,
                    'height': 1080,
                    'frame_rate': 23.976,
                },
                'HLS': {
                    'uri': 'v1/prog_index.m3u8'
                }
            }
        )
        t.tracks.append(vtrack)

        # add an audio track
        atrack = otio.schema.Track(
            "a1",
            kind=otio.schema.TrackKind.Audio
        )
        atrack.metadata.update(
            {
                'linked_tracks': [vtrack.name],
                'streaming': {
                    'bandwidth': 12345,
                    'codec': 'aac.test',
                    'group_id': 'aud1',
                },
                'HLS': {
                    'uri': 'a1/prog_index.m3u8'
                }
            }
        )
        t.tracks.append(atrack)

        # Write out and validate the playlist
        with tempfile.TemporaryDirectory() as temp_dir:
            media_pl_tmp_path = os.path.join(
                temp_dir,
                "master.m3u8"
            )
            otio.adapters.write_to_file(t, media_pl_tmp_path)

            with open(media_pl_tmp_path) as f:
                pl_string = f.read()

        # Drop blank lines before comparing
        pl_string = '\n'.join((line for line in pl_string.split('\n') if line))

        # Compare against the reference value
        self.assertEqual(pl_string, MEM_MASTER_PLAYLIST_REF_VALUE)

    def test_master_pl_with_iframe_pl_from_mem(self):
        t = otio.schema.Timeline()

        # add a video track
        vtrack = otio.schema.Track(
            "v1",
            kind=otio.schema.TrackKind.Video
        )
        vtrack.metadata.update(
            {
                'streaming': {
                    'bandwidth': 123456,
                    'codec': 'avc.test',
                    'width': 1920,
                    'height': 1080,
                    'frame_rate': 23.976,
                },
                'HLS': {
                    'uri': 'v1/prog_index.m3u8',
                    'iframe_uri': 'v1/iframe_index.m3u8'
                }
            }
        )
        t.tracks.append(vtrack)

        # add an audio track
        atrack = otio.schema.Track(
            "a1",
            kind=otio.schema.TrackKind.Audio
        )
        atrack.metadata.update(
            {
                'linked_tracks': [vtrack.name],
                'streaming': {
                    'bandwidth': 12345,
                    'codec': 'aac.test',
                    'group_id': 'aud1',
                },
                'HLS': {
                    'uri': 'a1/prog_index.m3u8'
                }
            }
        )
        t.tracks.append(atrack)

        # Write out and validate the playlist
        with tempfile.TemporaryDirectory() as temp_dir:
            media_pl_tmp_path = os.path.join(
                temp_dir,
                "master.m3u8"
            )
            otio.adapters.write_to_file(t, media_pl_tmp_path)

            with open(media_pl_tmp_path) as f:
                pl_string = f.read()

        # Drop blank lines before comparing
        pl_string = '\n'.join(line for line in pl_string.split('\n') if line)

        # Compare against the reference value
        self.assertEqual(pl_string, MEM_IFRAME_MASTER_PLAYLIST_REF_VALUE)

    def test_master_pl_complex_from_mem(self):
        t = otio.schema.Timeline()

        # add a video track
        vtrack = otio.schema.Track(
            "v1",
            kind=otio.schema.TrackKind.Video
        )
        vtrack.metadata.update(
            {
                'streaming': {
                    'bandwidth': 123456,
                    'codec': 'avc.test',
                    'width': 1920,
                    'height': 1080,
                    'frame_rate': 23.976,
                },
                'HLS': {
                    'uri': 'v1/prog_index.m3u8',
                    'iframe_uri': 'v1/iframe_index.m3u8'
                }
            }
        )
        t.tracks.append(vtrack)

        # add an alternate video track rep
        v2track = otio.schema.Track(
            "v2",
            kind=otio.schema.TrackKind.Video
        )
        v2track.metadata.update(
            {
                'streaming': {
                    'bandwidth': 12345,
                    'codec': 'avc.test',
                    'width': 720,
                    'height': 480,
                    'frame_rate': 23.976,
                },
                'HLS': {
                    'uri': 'v2/prog_index.m3u8',
                    'iframe_uri': 'v2/iframe_index.m3u8'
                }
            }
        )
        t.tracks.append(v2track)

        # add an audio track
        atrack = otio.schema.Track(
            "a1",
            kind=otio.schema.TrackKind.Audio
        )
        atrack.metadata.update(
            {
                'linked_tracks': [vtrack.name, v2track.name],
                'streaming': {
                    'bandwidth': 12345,
                    'codec': 'aac.test',
                    'group_id': 'aud1',
                },
                'HLS': {
                    'uri': 'a1/prog_index.m3u8'
                }
            }
        )
        t.tracks.append(atrack)

        # Write out and validate the playlist
        with tempfile.TemporaryDirectory() as temp_dir:
            media_pl_tmp_path = os.path.join(
                temp_dir,
                "master.m3u8"
            )
            otio.adapters.write_to_file(t, media_pl_tmp_path)

            with open(media_pl_tmp_path) as f:
                pl_string = f.read()

        # Drop blank lines before comparing
        pl_string = '\n'.join(line for line in pl_string.split('\n') if line)

        # Compare against the reference value
        self.assertEqual(pl_string, MEM_COMPLEX_MASTER_PLAYLIST_REF_VALUE)

    def test_master_playlist_hint_metadata(self):
        """
        Test that URL hints for master playlists don't leak out to media
        playlsits.
        """
        # Start with the reference playlist
        hls_path = HLS_EXAMPLE_PATH
        timeline = otio.adapters.read_from_file(hls_path)

        # add master playlist metadata to the track
        timeline.tracks[0].metadata.update(
            {
                'bandwidth': 123456,
                'codec': 'avc.test',
                'width': 1920,
                'height': 1080,
                'frame_rate': 23.976,
                'HLS': {
                    'uri': 'v1/prog_index.m3u8',
                    'iframe_uri': 'v1/iframe_index.m3u8'
                }
            }
        )

        # Write out and validate the playlist
        with tempfile.TemporaryDirectory() as temp_dir:
            media_pl_tmp_path = os.path.join(
                temp_dir,
                "test_media_pl_from_mem.m3u8"
            )
            otio.adapters.write_to_file(timeline, media_pl_tmp_path)

            with open(media_pl_tmp_path) as f:
                pl_string = f.read()

        # ensure metadata that wasn't supposed to didn't leak out
        for line in pl_string.split('\n'):
            self.assertFalse(line.startswith('#uri:'))
            self.assertFalse(line.startswith('#iframe_uri:'))

    def test_explicit_master_pl_from_mem(self):
        """Test that forcing a master playlist for a single track timeline
        works.
        """
        t = otio.schema.Timeline()
        # Set the master playlist flag
        t.metadata.update(
            {
                'HLS': {
                    'master_playlist': True
                }
            }
        )

        # build a track
        track = otio.schema.Track('v1')
        track.metadata.update(
            {
                'streaming': {
                    'bandwidth': 123456,
                    'codec': 'avc.test',
                    'width': 1920,
                    'height': 1080,
                    'frame_rate': 23.976,
                },
                'HLS': {
                    'EXT-X-INDEPENDENT-SEGMENTS': None,
                    'EXT-X-PLAYLIST-TYPE': 'VOD',
                    'uri': 'v1/prog_index.m3u8',
                    'iframe_uri': 'v1/iframe_index.m3u8'
                }
            }
        )
        t.tracks.append(track)

        # Make a prototype media ref with the segment's initialization metadata
        segmented_media_ref = otio.schema.ExternalReference(
            target_url='video1.mp4',
            metadata={
                'streaming': {
                    'init_byterange': {
                        'byte_count': 729,
                        'byte_offset': 0
                    },
                    'init_uri': 'media-video-1.mp4'
                }
            }
        )

        # Make a copy of the media ref specifying the byte range for the
        # segment
        media_ref1 = segmented_media_ref.deepcopy()
        media_ref1.available_range = otio.opentime.TimeRange(
            otio.opentime.RationalTime(0, 1),
            otio.opentime.RationalTime(2.002, 1)
        )
        media_ref1.metadata.update(
            {
                'streaming': {
                    'byte_count': 534220,
                    'byte_offset': 1361
                }
            }
        )

        # make the segment and append it
        segment1 = otio.schema.Clip(media_reference=media_ref1)
        track.append(segment1)

        # Write out and validate the playlist
        with tempfile.TemporaryDirectory() as temp_dir:
            master_pl_tmp_path = os.path.join(
                temp_dir,
                "master.m3u8"
            )
            otio.adapters.write_to_file(t, master_pl_tmp_path)

            with open(master_pl_tmp_path) as f:
                pl_string = f.read()

        # Drop blank lines before comparing
        pl_string = '\n'.join((line for line in pl_string.split('\n') if line))

        # Compare against the reference value
        self.assertEqual(pl_string, MEM_SINGLE_TRACK_MASTER_PLAYLIST_REF_VALUE)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Test builtin adapters."""

import os
import unittest

import opentimelineio as otio
import opentimelineio.test_utils as otio_test_utils

from opentimelineio.adapters import (
    otio_json,
)

# handle python2 vs python3 difference
try:
    from tempfile import TemporaryDirectory  # noqa: F401
    import tempfile
except ImportError:
    # XXX: python2.7 only
    from backports import tempfile


SAMPLE_DATA_DIR = os.path.join(os.path.dirname(__file__), "sample_data")
SCREENING_EXAMPLE_PATH = os.path.join(SAMPLE_DATA_DIR, "screening_example.edl")


class BuiltInAdapterTest(unittest.TestCase, otio_test_utils.OTIOAssertions):

    def test_disk_io(self):
        edl_path = SCREENING_EXAMPLE_PATH
        timeline = otio.adapters.read_from_file(edl_path)

        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = os.path.join(temp_dir, "test_disk_io.otio")
            otio.adapters.write_to_file(timeline, temp_file)
            decoded = otio.adapters.read_from_file(temp_file)
            self.assertJsonEqual(timeline, decoded)

    def test_otio_round_trip(self):
        tl = otio.adapters.read_from_file(SCREENING_EXAMPLE_PATH)

        baseline_json = otio.adapters.otio_json.write_to_string(tl)

        self.assertEqual(tl.name, "Example_Screening.01")

        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = os.path.join(temp_dir, 'test_otio_round_trip.otio')
            otio.adapters.otio_json.write_to_file(tl, temp_file)
            new = otio.adapters.otio_json.read_from_file(temp_file)

            new_json = otio.adapters.otio_json.write_to_string(new)

            self.assertMultiLineEqual(baseline_json, new_json)
            self.assertIsOTIOEquivalentTo(tl, new)

    def test_disk_vs_string(self):
        """ Writing to disk and writing to a string should
        produce the same result
        """
        timeline = otio.adapters.read_from_file(SCREENING_EXAMPLE_PATH)

        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = os.path.join(temp_dir, "test_disk_vs_string.otio")
            otio.adapters.write_to_file(timeline, temp_file)
            in_memory = otio.adapters.write_to_string(timeline, 'otio_json')
            with open(temp_file, 'r') as f:
                on_disk = f.read()

            self.maxDiff = None

            # for debugging
            # with open("/var/tmp/in_memory.otio", "w") as fo:
            #     fo.write(in_memory)
            #
            # with open("/var/tmp/on_disk.otio", "w") as fo:
            #     fo.write(on_disk)

            self.assertEqual(in_memory, on_disk)

    def test_adapters_fetch(self):
        """ Test the dynamic string based adapter fetching """
        self.assertEqual(
            otio.adapters.from_name('otio_json').module(),
            otio_json
        )

    def test_otio_json_default(self):
        tl = otio.adapters.read_from_file(SCREENING_EXAMPLE_PATH)
        self.assertMultiLineEqual(
            otio.adapters.write_to_string(tl, 'otio_json'),
            otio.adapters.write_to_string(tl)
        )

        test_str = otio.adapters.write_to_string(tl)
        self.assertJsonEqual(tl, otio.adapters.read_from_string(test_str))


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

import unittest

import opentimelineio as otio
import opentimelineio.test_utils as otio_test_utils


class Box2dTests(unittest.TestCase, otio_test_utils.OTIOAssertions):
    def test_cons(self):
        v1 = otio.schema.V2d(1.0, 2.0)
        v2 = otio.schema.V2d(3.0, 4.0)
        box = otio.schema.Box2d(v1, v2)

        self.assertEqual(box.min, v1)
        self.assertEqual(box.max, v2)

    def test_str(self):
        v1 = otio.schema.V2d(1.0, 2.0)
        v2 = otio.schema.V2d(3.0, 4.0)
        box = otio.schema.Box2d(v1, v2)

        self.assertMultiLineEqual(
            str(box),
            'Box2d(V2d(1.0, 2.0), V2d(3.0, 4.0))'
        )

        self.assertMultiLineEqual(
            repr(box),
            'otio.schema.Box2d('
            'min=otio.schema.V2d(x=1.0, y=2.0), '
            'max=otio.schema.V2d(x=3.0, y=4.0)'
            ')'
        )

    def test_center(self):
        v1 = otio.schema.V2d(1.0, 2.0)
        v2 = otio.schema.V2d(3.0, 4.0)
        box1 = otio.schema.Box2d(v1, v2)

        self.assertEqual(box1.center(),
                         otio.schema.V2d(2.0, 3.0)
                         )

    def test_extend(self):
        v1 = otio.schema.V2d(1.0, 2.0)
        v2 = otio.schema.V2d(3.0, 4.0)
        box1 = otio.schema.Box2d(v1, v2)

        box1.extendBy(otio.schema.V2d(5.0, 5.0))
        self.assertEqual(box1,
                         otio.schema.Box2d(
                             otio.schema.V2d(1.0, 2.0),
                             otio.schema.V2d(5.0, 5.0)
                         )
                         )

        v3 = otio.schema.V2d(2.0, 3.0)
        v4 = otio.schema.V2d(6.0, 6.0)
        box2 = otio.schema.Box2d(v3, v4)

        box1.extendBy(box2)
        self.assertEqual(box1,
                         otio.schema.Box2d(
                             otio.schema.V2d(1.0, 2.0),
                             otio.schema.V2d(6.0, 6.0)
                         )
                         )

    def test_intersects(self):
        v1 = otio.schema.V2d(1.0, 2.0)
        v2 = otio.schema.V2d(3.0, 4.0)
        box1 = otio.schema.Box2d(v1, v2)

        self.assertTrue(box1.intersects(otio.schema.V2d(2.0, 3.0)))

        self.assertFalse(box1.intersects(otio.schema.V2d(0.0, 3.0)))

        v3 = otio.schema.V2d(1.1, 1.9)
        v4 = otio.schema.V2d(3.1, 3.9)
        box2 = otio.schema.Box2d(v3, v4)

        self.assertTrue(box1.intersects(box2))

        v5 = otio.schema.V2d(3.1, 4.1)
        v6 = otio.schema.V2d(4.1, 5.1)
        box3 = otio.schema.Box2d(v5, v6)

        self.assertFalse(box1.intersects(box3))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

import math
import os
import sys
import unittest

# handle python2 vs python3 difference
try:
    from tempfile import TemporaryDirectory  # noqa: F401
    import tempfile
except ImportError:
    # XXX: python2.7 only
    from backports import tempfile

import opentimelineio as otio
import opentimelineio.test_utils as otio_test_utils

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(__file__), "sample_data")
BIG_INT_TEST = os.path.join(SAMPLE_DATA_DIR, "big_int.otio")


class TimelineTests(unittest.TestCase, otio_test_utils.OTIOAssertions):

    def test_init(self):
        rt = otio.opentime.RationalTime(12, 24)
        tl = otio.schema.Timeline("test_timeline", global_start_time=rt)
        self.assertEqual(tl.name, "test_timeline")
        self.assertEqual(tl.global_start_time, rt)

    def test_metadata(self):
        rt = otio.opentime.RationalTime(12, 24)
        tl = otio.schema.Timeline("test_timeline", global_start_time=rt)
        tl.metadata['foo'] = "bar"
        self.assertEqual(tl.metadata['foo'], 'bar')

        encoded = otio.adapters.otio_json.write_to_string(tl)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(tl, decoded)
        self.assertEqual(tl.metadata, decoded.metadata)

    def test_big_integers(self):
        result = otio.adapters.read_from_file(BIG_INT_TEST)

        md = result.tracks[0][0].metadata["int_test"]

        # test the deserialized values from the file

        # positive integers
        self.assertEqual(md["maxint32"], 2147483647)
        self.assertEqual(md["smallest_int64"], 2147483648)
        self.assertEqual(md["verybig"], 3450100000)

        # negative
        self.assertEqual(md["minint32"], -2147483647)
        self.assertEqual(md["neg_smallest_int64"], -2147483648)
        self.assertEqual(md["negverybig"], -3450100000)

        big_int_type = int
        if sys.version_info[0] < 3:
            big_int_type = long  # noqa: F821

        # from memory
        supported_integers = [
            # name       value to enter
            ('minint32', -big_int_type(2**31 - 1)),
            ('maxint32', big_int_type(2**31 - 1)),
            ('maxuint32', big_int_type(2**32 - 1)),
            ('minint64', -big_int_type(2**63 - 1)),
            ('maxint64', big_int_type(2**63 - 1)),
        ]

        for (name, value) in supported_integers:
            md[name] = value
            self.assertEqual(
                md[name],
                value,
                "{} didn't match expected value: got {} expected {}".format(
                    name,
                    md[name],
                    value
                )
            )
            self.assertEqual(
                type(md[name]),
                big_int_type,
                "{} didn't match expected type: got {} expected {}".format(
                    name,
                    type(md[name]),
                    big_int_type
                )
            )

            # test that roundtripping through json functions
            serialized = otio.adapters.write_to_string(
                result,
                "otio_json"
            )
            deserialized = otio.adapters.read_from_string(
                serialized,
                "otio_json"
            )

            value_deserialized = (
                deserialized.tracks[0][0].metadata["int_test"][name]
            )
            self.assertEqual(
                value,
                value_deserialized,
                "{} did not round trip correctly.  expected: {} got: {}".format(
                    name,
                    value,
                    value_deserialized
                )
            )
            self.assertEqual(
                type(value),
                type(value_deserialized),
                "the type of {} did not round trip correctly.  expected: "
                "{} got: {}".format(
                    name,
                    type(value),
                    type(value_deserialized)
                )
            )

        supported_doubles = [
            # name       value to enter
            ('minint32', -float(2**31 - 1)),
            ('maxint32', float(2**31 - 1)),
            ('maxuint32', float(2**32 - 1)),
            ('minint64', -float(2**63 - 1)),
            ('maxint64', float(2**63 - 1)),
            ('maxdouble', sys.float_info.max),
            ('infinity', float('inf')),
            ('infinity_because_too_big', float(2 * sys.float_info.max)),
            ('nan', float('nan')),
            ('neg_infinity', float('-inf')),
        ]

        for (name, value) in supported_doubles:
            md[name] = value

            # float('nan') != float('nan'), so need ot test isnan(value) instead
            if not math.isnan(value):
                self.assertEqual(
                    md[name],
                    value,
                    "{} didn't match expected value: got '{}' ('{}') expected "
                    "'{}' ('{}')".format(
                        name,
                        md[name],
                        type(md[name]),
                        value,
                        type(value)
                    )
                )
            else:
                self.assertTrue(
                    math.isnan(md[name]),
                    "Expected {} to be a nan, got {}".format(name, md[name])
                )

            self.assertEqual(
                type(md[name]),
                float,
                "{} didn't match expected type: got {} expected {}".format(
                    name,
                    type(md[name]),
                    float
                )
            )

            # test that roundtripping through json functions
            try:
                serialized = otio.adapters.write_to_string(
                    result,
                    "otio_json"
                )
            except Exception as e:
                self.fail(
                    "A problem occured when attempting to serialize {} "
                    "with value: {}.  Error message was: {}".format(
                        name,
                        value,
                        e
                    )
                )

            try:
                deserialized = otio.adapters.read_from_string(
                    serialized,
                    "otio_json"
                )
            except Exception as e:
                self.fail(
                    "A problem occured when attempting to serialize {} "
                    "with value: {}.  Error message was: {}".format(
                        name,
                        value,
                        e
                    )
                )

            value_deserialized = (
                deserialized.tracks[0][0].metadata["int_test"][name]
            )

            if not math.isnan(value):
                self.assertEqual(
                    value,
                    value_deserialized,
                    "{} did not round trip correctly.  expected: {}, of type {} "
                    "got: {}, of type {}".format(
                        name,
                        value,
                        type(value),
                        value_deserialized,
                        type(value_deserialized),
                    )
                )
            else:
                self.assertTrue(
                    math.isnan(value_deserialized),
                    "Expected {} to be a nan, got {}".format(
                        name,
                        value_deserialized
                    )
                )

        unsupported_values = [
            # numbers -- supported python type but don't fit into the C++ types
            #            (ie int but doesn't fit into int64_t)
            ('minuint64_not_int64', int(2**63), ValueError),
            ('maxuint64', int(2**64 - 1), ValueError),
            ('minint128', int(2**64), ValueError),

            # other kinds of python things
            ('object', object(), TypeError),
            ('set', set(), TypeError),
        ]

        for (name, value, exc) in unsupported_values:
            with self.assertRaises(
                    exc,
                    msg="Expected {} to raise an exception.".format(name)
            ):
                md[name] = value

    def test_unicode(self):
        result = otio.adapters.read_from_file(BIG_INT_TEST)

        md = result.tracks[0][0].metadata['unicode']

        utf8_test_str = "Viel glück und hab spaß!"

        # python2
        if sys.version_info[0] < 3:
            utf8_test_str = utf8_test_str.decode('utf8')

        self.assertEqual(md['utf8'], utf8_test_str)

        tl = otio.schema.Timeline()

        tl.metadata['utf8'] = utf8_test_str
        self.assertEqual(tl.metadata['utf8'], utf8_test_str)

        encoded = otio.adapters.otio_json.write_to_string(tl)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(tl, decoded)
        self.assertEqual(tl.metadata, decoded.metadata)

    @unittest.skipIf(sys.version_info < (3, 0), "unicode does funny things in python2")
    def test_unicode_file_name(self):
        with tempfile.TemporaryDirectory() as temp_dir:

            tl = otio.schema.Timeline("大平原")
            filename = os.path.join(temp_dir, "大平原.otio")
            otio.adapters.write_to_file(tl, filename)

            result = otio.adapters.read_from_file(filename)
            self.assertEqual(result.name, "大平原")

    def test_range(self):
        track = otio.schema.Track(name="test_track")
        tl = otio.schema.Timeline("test_timeline", tracks=[track])
        rt = otio.opentime.RationalTime(5, 24)
        mr = otio.schema.ExternalReference(
            available_range=otio.opentime.range_from_start_end_time(
                otio.opentime.RationalTime(5, 24),
                otio.opentime.RationalTime(15, 24)
            ),
            target_url="/var/tmp/test.mov"
        )

        cl = otio.schema.Clip(
            name="test clip1",
            media_reference=mr,
            source_range=otio.opentime.TimeRange(duration=rt),
        )
        cl2 = otio.schema.Clip(
            name="test clip2",
            media_reference=mr,
            source_range=otio.opentime.TimeRange(duration=rt),
        )
        cl3 = otio.schema.Clip(
            name="test clip3",
            media_reference=mr,
            source_range=otio.opentime.TimeRange(duration=rt),
        )
        tl.tracks[0].append(cl)
        tl.tracks[0].extend([cl2, cl3])

        self.assertEqual(tl.duration(), rt + rt + rt)

        self.assertEqual(
            tl.range_of_child(cl),
            tl.tracks[0].range_of_child_at_index(0)
        )

    def test_available_image_bounds(self):
        track = otio.schema.Track(name="test_track")
        tl = otio.schema.Timeline("test_timeline", tracks=[track])

        # three clips, each successive clip partially overlaps the previous
        cl = otio.schema.Clip(
            name="test clip1",
            media_reference=otio.schema.ExternalReference(
                available_image_bounds=otio.schema.Box2d(
                    otio.schema.V2d(0.0, 0.0),
                    otio.schema.V2d(1.0, 1.0)
                ),
                target_url="/var/tmp/test.mov"
            )
        )
        cl2 = otio.schema.Clip(
            name="test clip2",
            media_reference=otio.schema.ExternalReference(
                available_image_bounds=otio.schema.Box2d(
                    otio.schema.V2d(1.0, 1.0),
                    otio.schema.V2d(2.0, 2.0)
                ),
                target_url="/var/tmp/test.mov"
            )
        )
        cl3 = otio.schema.Clip(
            name="test clip3",
            media_reference=otio.schema.ExternalReference(
                available_image_bounds=otio.schema.Box2d(
                    otio.schema.V2d(2.0, 2.0),
                    otio.schema.V2d(3.0, 3.0)
                ),
                target_url="/var/tmp/test.mov"
            )
        )
        gap = otio.schema.Gap(name="gap")

        tl.tracks[0].append(cl)
        tl.tracks[0].extend([cl2, cl3, gap])

        union = otio.schema.Box2d(
            otio.schema.V2d(0.0, 0.0),
            otio.schema.V2d(3.0, 3.0)
        )

        # union should be overlapping area, gap should be ignored
        self.assertEqual(tl.tracks[0].available_image_bounds, union)

    def test_iterators(self):
        self.maxDiff = None
        track = otio.schema.Track(name="test_track")
        tl = otio.schema.Timeline("test_timeline", tracks=[track])
        rt = otio.opentime.RationalTime(5, 24)
        mr = otio.schema.ExternalReference(
            available_range=otio.opentime.range_from_start_end_time(
                otio.opentime.RationalTime(5, 24),
                otio.opentime.RationalTime(15, 24)
            ),
            target_url="/var/tmp/test.mov"
        )

        cl = otio.schema.Clip(
            name="test clip1",
            media_reference=mr,
            source_range=otio.opentime.TimeRange(
                mr.available_range.start_time,
                rt
            ),
        )
        cl2 = otio.schema.Clip(
            name="test clip2",
            media_reference=mr,
            source_range=otio.opentime.TimeRange(
                mr.available_range.start_time,
                rt
            ),
        )
        cl3 = otio.schema.Clip(
            name="test clip3",
            media_reference=mr,
            source_range=otio.opentime.TimeRange(
                mr.available_range.start_time,
                rt
            ),
        )
        tl.tracks[0].append(cl)
        tl.tracks[0].extend([cl2, cl3])
        self.assertEqual([cl, cl2, cl3], list(tl.each_clip()))

        rt_start = otio.opentime.RationalTime(0, 24)
        rt_end = otio.opentime.RationalTime(1, 24)
        search_range = otio.opentime.TimeRange(rt_start, rt_end)
        self.assertEqual([cl], list(tl.each_clip(search_range)))

        # check to see if full range works
        search_range = tl.tracks.trimmed_range()
        self.assertEqual([cl, cl2, cl3], list(tl.each_clip(search_range)))

        # just one clip
        search_range = cl2.range_in_parent()
        self.assertEqual([cl2], list(tl.each_clip(search_range)))

        # the last two clips
        search_range = otio.opentime.TimeRange(
            start_time=cl2.range_in_parent().start_time,
            duration=cl2.trimmed_range().duration + rt_end
        )
        self.assertEqual([cl2, cl3], list(tl.each_clip(search_range)))

        # no clips
        search_range = otio.opentime.TimeRange(
            start_time=otio.opentime.RationalTime(
                value=-10,
                rate=rt_start.rate
            ),
            duration=rt_end
        )
        self.assertEqual([], list(tl.each_clip(search_range)))

    def test_str(self):
        self.maxDiff = None
        clip = otio.schema.Clip(
            name="test_clip",
            media_reference=otio.schema.MissingReference()
        )
        track = otio.schema.Track(name="test_track", children=[clip])
        tl = otio.schema.Timeline(name="test_timeline", tracks=[track])
        self.assertMultiLineEqual(
            str(tl),
            'Timeline(' +
            '"' + str(tl.name) + '", ' +
            str(tl.tracks) +
            ')'
        )
        self.assertMultiLineEqual(
            repr(tl),
            'otio.schema.Timeline(' +
            "name='" + tl.name + "', " +
            "tracks=" + repr(tl.tracks) +
            ')'
        )

    def test_serialize_timeline(self):
        clip = otio.schema.Clip(
            name="test_clip",
            media_reference=otio.schema.MissingReference()
        )
        tl = otio.schema.timeline_from_clips([clip])
        encoded = otio.adapters.otio_json.write_to_string(tl)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(tl, decoded)

        string2 = otio.adapters.otio_json.write_to_string(decoded)
        self.assertEqual(encoded, string2)

    def test_serialization_of_subclasses(self):
        clip1 = otio.schema.Clip()
        clip1.name = "Test Clip"
        clip1.media_reference = otio.schema.ExternalReference(
            "/tmp/foo.mov"
        )
        tl1 = otio.schema.timeline_from_clips([clip1])
        tl1.name = "Testing Serialization"
        self.assertIsNotNone(tl1)
        otio_module = otio.adapters.from_name("otio_json")
        serialized = otio_module.write_to_string(tl1)
        self.assertIsNotNone(serialized)
        tl2 = otio_module.read_from_string(serialized)
        self.assertIsNotNone(tl2)
        self.assertEqual(type(tl1), type(tl2))
        self.assertEqual(tl1.name, tl2.name)
        self.assertEqual(len(tl1.tracks), 1)
        self.assertEqual(len(tl2.tracks), 1)
        track1 = tl1.tracks[0]
        track2 = tl2.tracks[0]
        self.assertEqual(type(track1), type(track2))
        self.assertEqual(len(track1), 1)
        self.assertEqual(len(track2), 1)
        clip2 = tl2.tracks[0][0]
        self.assertEqual(clip1.name, clip2.name)
        self.assertEqual(type(clip1), type(clip2))
        self.assertEqual(
            type(clip1.media_reference),
            type(clip2.media_reference)
        )
        self.assertEqual(
            clip1.media_reference.target_url,
            clip2.media_reference.target_url
        )

    def test_tracks(self):
        tl = otio.schema.Timeline(tracks=[
            otio.schema.Track(
                name="V1",
                kind=otio.schema.TrackKind.Video
            ),
            otio.schema.Track(
                name="V2",
                kind=otio.schema.TrackKind.Video
            ),
            otio.schema.Track(
                name="A1",
                kind=otio.schema.TrackKind.Audio
            ),
            otio.schema.Track(
                name="A2",
                kind=otio.schema.TrackKind.Audio
            ),
        ])
        self.assertListEqual(
            ["V1", "V2"],
            [t.name for t in tl.video_tracks()]
        )
        self.assertListEqual(
            ["A1", "A2"],
            [t.name for t in tl.audio_tracks()]
        )

    def test_tracks_set_null_tracks(self):
        tl = otio.schema.Timeline(tracks=[
            otio.schema.Track(
                name="V1",
                kind=otio.schema.TrackKind.Video
            ),
            otio.schema.Track(
                name="V2",
                kind=otio.schema.TrackKind.Video
            )])

        self.assertEqual(len(tl.tracks), 2)
        self.assertTrue(isinstance(tl.tracks, otio.schema.Stack))
        tl.tracks = None
        self.assertEqual(len(tl.audio_tracks()), 0)
        self.assertEqual(len(tl.video_tracks()), 0)
        self.assertTrue(isinstance(tl.tracks, otio.schema.Stack))

    def test_children_if(self):
        cl = otio.schema.Clip()
        tr = otio.schema.Track()
        tr.append(cl)
        tl = otio.schema.Timeline()
        tl.tracks.append(tr)
        result = tl.children_if(otio.schema.Clip)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], cl)

    def test_children_if_search_range(self):
        range = otio.opentime.TimeRange(
            otio.opentime.RationalTime(0.0, 24.0),
            otio.opentime.RationalTime(24.0, 24.0))
        cl0 = otio.schema.Clip()
        cl0.source_range = range
        cl1 = otio.schema.Clip()
        cl1.source_range = range
        cl2 = otio.schema.Clip()
        cl2.source_range = range
        tr = otio.schema.Track()
        tr.append(cl0)
        tr.append(cl1)
        tr.append(cl2)
        tl = otio.schema.Timeline()
        tl.tracks.append(tr)
        result = tl.children_if(otio.schema.Clip, range)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], cl0)

    def test_children_if_shallow_search(self):
        cl = otio.schema.Clip()
        tr = otio.schema.Track()
        tr.append(cl)
        tl = otio.schema.Timeline()
        tl.tracks.append(tr)
        result = tl.children_if(otio.schema.Clip, shallow_search=True)
        self.assertEqual(len(result), 0)
        result = tl.children_if(otio.schema.Clip, shallow_search=False)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], cl)


if __name__ == '__main__':
    unittest.main()
